import os
import re
import unittest
from puma.utils import PROJECT_ROOT
from puma.version import version as setup_version


def read_release_notes(file_path: str) -> [str]:
    """
    Read the content of the release notes file.
    :param file_path: The path to the release notes file.
    :return: The content of the release notes file.
    """
    with open(file_path, 'r') as file:
        return file.readlines()


class TestVersion(unittest.TestCase):

    def setUp(self):
        self.release_notes_path = f"{PROJECT_ROOT}/RELEASE_NOTES"
        self.release_notes = read_release_notes(self.release_notes_path)
        self.first_version_in_release_notes = self._get_first_version_in_release_notes()

    def _get_first_version_in_release_notes(self):
        first_line = self.release_notes[0]
        match = re.search(r'(\d+\.\d+\.\d+)', first_line)
        return match.group(1) if match else None

    def test_version_in_release_notes_same_as_setup(self):
        self.assertIsNotNone(self.first_version_in_release_notes)
        self.assertEqual(self.first_version_in_release_notes, setup_version,
                         "Version in release notes is not equal to setup version")

    def test_versions_same_as_github(self):
        github_tag_version = os.getenv('GITHUB_TAG_VERSION')
        # Only run this test when a release is made from GitHub (i.e. the above environment variable is set)
        if github_tag_version is None:
            self.skipTest("Skipping GitHub version test as no tag version was passed.")

        self.assertEqual(github_tag_version, self.first_version_in_release_notes,
                         "GitHub tag version is not equal to top version in release notes")
        self.assertEqual(github_tag_version, setup_version,
                         "GitHub tag version is not equal to setup version")



from enum import Enum
from time import sleep
from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android.appium_actions import AndroidAppiumActions, supported_version
from puma.utils.route_simulator import RouteSimulator

GOOGLE_MAPS_PACKAGE = 'com.google.android.apps.maps'


class TransportType(Enum):
    CAR = "car"
    BIKE = "bike"


@supported_version("11.119.0101")
class GoogleMapsActions(AndroidAppiumActions):
    def __init__(self,
                 device_udid,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait=1,
                 appium_server='http://localhost:4723'):
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      GOOGLE_MAPS_PACKAGE,
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=implicit_wait,
                                      appium_server=appium_server)
        self.route_simulator = RouteSimulator(self, 0)

    def _has_search_box(self):
        return self.app_open() and self.is_present('//android.widget.TextView[@text="Search here"]')

    def _ensure_at_start(self):
        while not self._has_search_box():
            self.driver.back()
            if not self.app_open():
                self.activate_app()

    def search_place(self, search_string: str):
        self._ensure_at_start()
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.TextView[@text="Search here"]').click()
        search_box_xpath = '//android.widget.EditText[@resource-id="com.google.android.apps.maps:id/search_omnibox_edit_text"]'
        self.driver.find_element(by=AppiumBy.XPATH, value=search_box_xpath).send_keys(search_string)
        first_result = '//android.support.v7.widget.RecyclerView[@resource-id="com.google.android.apps.maps:id/typed_suggest_container"]/android.widget.LinearLayout[1]'
        self.driver.find_element(by=AppiumBy.XPATH, value=first_result).click()

    def start_navigation(self, search_string: str, transport_type: TransportType = TransportType.CAR, time_to_wait=10):
        self.search_place(search_string)
        directions_xpath = '//android.widget.Button[starts-with(@content-desc, "Directions to")]'
        self.driver.find_element(by=AppiumBy.XPATH, value=directions_xpath).click()
        if transport_type == TransportType.CAR:
            mode = 'Driving mode'
        elif transport_type == TransportType.BIKE:
            mode = 'Bicycling mode'
        multiple_choices_header = '//android.widget.LinearLayout[@resource-id="com.google.android.apps.maps:id/destination_list_header"]'
        if self.is_present(multiple_choices_header):
            # select first option
            first_option = '//android.support.v7.widget.RecyclerView[@resource-id="com.google.android.apps.maps:id/recycler_view"]/android.widget.LinearLayout[1]'
            self.driver.find_element(by=AppiumBy.XPATH, value=first_option).click()
        mode_xpath = f'//android.widget.LinearLayout[starts-with(@content-desc, "{mode}")]'
        self.driver.find_element(by=AppiumBy.XPATH, value=mode_xpath).click()
        start_xpath = '//android.widget.Button[@content-desc="Start driving navigation"]'
        for i in range(time_to_wait):
            if self.is_present(start_xpath):
                self.driver.find_element(by=AppiumBy.XPATH, value=start_xpath).click()
                return
            else:
                sleep(1)
        raise Exception(f'Route was not loaded after {time_to_wait} seconds')

    def start_route(self, from_query: str, to_query: str, speed: int,
                    transport_type: TransportType = TransportType.CAR):
        self.route_simulator.update_speed(0)
        osm_mode = 'car' if transport_type == TransportType.CAR else 'bike'
        self.route_simulator.execute_route_with_queries(from_query, to_query, osm_mode)
        self.start_navigation(to_query, transport_type)
        self.route_simulator.update_speed(speed)



from time import sleep
from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android.appium_actions import AndroidAppiumActions, supported_version

SNAPCHAT_PACKAGE = 'com.snapchat.android'


@supported_version("12.90.0.46")
class SnapchatActions(AndroidAppiumActions):
    def __init__(self,
                 device_udid,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait=1,
                 appium_server='http://localhost:4723'):
        """
        Class with an API for Snapchat Android using Appium. Can be used with an emulator or real device attached to the computer.
        """
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      SNAPCHAT_PACKAGE,
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=implicit_wait,
                                      appium_server=appium_server)

    def _currently_at_homescreen(self) -> bool:
        return self.is_present(
            '//android.widget.LinearLayout[@resource-id="com.snapchat.android:id/ngs_navigation_bar"]') \
            and not self.is_present('//*[@text="View Profile"]')

    def _currently_in_conversation_overview(self) -> bool:
        return self.is_present(
            '//android.widget.TextView[@resource-id="com.snapchat.android:id/hova_page_title" and @text="Chat"]')

    def _currently_on_top_of_conversation_overview(self) -> bool:
        """
        Check if the current position is at the top of the conversation view.
        """
        on_top = False
        header_patterns = ["Unread", "Groups", "Unreplied"]
        for pattern in header_patterns:
            if self.is_present(f'//android.widget.TextView[lower-case(@content-desc)="{pattern.lower()}"]'):
                on_top = True
                break
        return on_top

    def _currently_in_conversation(self) -> bool:
        return self.is_present(
            '//android.widget.EditText[@resource-id="com.snapchat.android:id/chat_input_text_field"]')

    def _currently_in_camera_tab(self) -> bool:
        """
        Check if the current position is the camera tab.
        The camera button is checked, but this element also occurs in the camera view when sending a photo to a contact.
        Thus, an additional check is needed if "Send To" is not present is required.
        :return:
        """
        return (
                self.is_present('//android.widget.FrameLayout[@content-desc="Camera Capture"]')
                and not self.is_present('//android.widget.TextView[@text="Send To"]')
        )

    def _go_to_main_tab(self, tab_name: str):
        """
        Navigate to one of the main tabs.
        :param tab_name: One of the main tabs of snapchat: Map, Chat, Camera, Stories, or Spotlight
        """
        if self.driver.current_package != SNAPCHAT_PACKAGE:
            self.driver.activate_app(SNAPCHAT_PACKAGE)
        while not self._currently_at_homescreen():
            self.driver.back()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value=f'//android.widget.LinearLayout[@resource-id="com.snapchat.android:id/ngs_navigation_bar"]/android.view.ViewGroup[@content-desc="{tab_name}"]').click()

    # TODO create a separate function for each tab, so the user cannot make any typos
    def go_to_conversation_tab(self):
        """
        Navigate to the top of the conversation tab.
        When tapping "navigate to chat" when already in the conversation
        tab, the focus shifts down to the Quick Add section. This poses a problem when selecting a specific
        conversation that is out of view. This method makes sure to be at the top
        """
        if not self._currently_in_conversation_overview():
            self._go_to_main_tab("Chat")
        if not self._currently_on_top_of_conversation_overview():
            self._go_to_main_tab("Chat")

    def go_to_camera_tab(self):
        """
        Navigate to camera tab.
        """
        if not self._currently_in_camera_tab():
            self._go_to_main_tab("Camera")

    def select_chat(self, chat_subject: str):
        """
        Opens a given conversation.
        :param chat_subject: the name of the conversation to open
        """
        self.go_to_conversation_tab()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value=f'//javaClass[contains(lower-case(@text), "{chat_subject.lower()}")]/..').click()

    def _if_chat_go_to_chat(self, chat: str):
        if chat is not None:
            self.select_chat(chat)
        sleep(1)
        if not self._currently_in_conversation():
            raise Exception('Expected to be in conversation screen now, but screen contents are unknown')

    def send_message(self, message: str, chat: str = None):
        """
        Sends a text message, either in the current conversation, or in a given conversation.
        :param message: the message to send
        :param chat: optional: the conversation in which to send the message. If not used, it is assumed the
                     conversation is already opened.
        """
        self._if_chat_go_to_chat(chat)
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.EditText[@resource-id="com.snapchat.android:id/chat_input_text_field"]') \
            .send_keys(message)
        self._press_enter()

    def _press_enter(self):
        enter_keycode = 66
        self.driver.press_keycode(enter_keycode)

    def send_snap(self, recipients: [str] = None, caption: str = None, front_camera: bool = True):
        """
        Sends a snap (picture), either to one or more contacts, or posts it to `My story`
        :param recipients: Optional: a list of recipients to send the snap to
        :param caption: Optional: a caption to set on the snap
        :param front_camera: Optional: whether or not to use the front camera (True by default)
        """
        # go to camera and snap picture
        self.go_to_camera_tab()
        if not front_camera:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.view.ViewGroup[@content-desc="Flip Camera"]').click()
            sleep(0.5)
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.FrameLayout[@content-desc="Camera Capture"]').click()
        # write caption if needed
        if caption:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.view.View[@resource-id="com.snapchat.android:id/full_screen_surface_view"]').click()
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.EditText[@resource-id="com.snapchat.android:id/caption_edit_text_view"]').send_keys(
                caption)
            self.driver.back()
        # press send
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.ImageButton[lower-case(@content-desc)="send"]').click()

        # select recipients, or post as story, and send
        if recipients:
            for recipient in recipients:
                self.driver.find_element(by=AppiumBy.XPATH,
                                         value=f'//javaClass[lower-case(@text) ="{recipient.lower()}"]').click()
        else:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//javaClass[lower-case(@text)="my story · friends only"]/..').click()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.view.View[lower-case(@content-desc)="send"]').click()



import os
import time
from datetime import datetime
from pathlib import Path
from typing import Dict
from uuid import uuid4

from adb_pywrapper.adb_device import AdbDevice
from adb_pywrapper.adb_screen_recorder import AdbScreenRecorder
from appium import webdriver
from appium.options.android import UiAutomator2Options
from appium.webdriver import WebElement
from appium.webdriver.common.appiumby import AppiumBy
from appium.webdriver.extensions.android.nativekey import AndroidKey
from appium.webdriver.webdriver import WebDriver
from selenium.common import NoSuchElementException
from urllib3.exceptions import MaxRetryError

from puma.apps.android import logger
from puma.computer_vision import ocr
from puma.utils.video_utils import CACHE_FOLDER, log_error_and_raise_exception

__drivers: dict[str, WebDriver] = {}

def _get_appium_driver(appium_server: str, udid: str, options) -> WebDriver:
    key = f"{appium_server}${udid}"
    if key not in __drivers.keys():
        try:
            __drivers[key] = webdriver.Remote(appium_server, options=options)
        except MaxRetryError:
            logger.error("Connecting to the Appium server has failed.\n"
                  "Make sure that the appium server is running!\n"
                  "This can be done by running the `appium` command from the command line.")
            exit(1)
    else:
        logger.warning(f'WARNING: there already was an initialized driver for appium server {appium_server} and udid {udid}. '
              'This driver will be used, which might mean your appium capabilities are ignored as these cannot be'
              'altered for a driver that has already been initialized. If you need specific capabilities, please '
              'rewrite your Puma code to ensure the correct capabilities are loaded the first time you connect to '
              f'server {appium_server} and device {udid}.')
    return __drivers[key]


def _get_android_default_options():
    options = UiAutomator2Options()
    options.no_reset = True
    options.platform_name = 'Android'
    options.new_command_timeout = 1200
    return options


def supported_version(version: str):
    def decorator(class_or_function):
        class_or_function.supported_version = version
        return class_or_function

    return decorator


class AndroidAppiumActions:

    def __init__(self,
                 udid: str,
                 app_package: str,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait: int = 1,
                 appium_server: str = 'http://localhost:4723'):
        """
        Class with a generic API for Appium scripting on an Android device.
        Can be used with an emulator or real device attached to the computer.
        :param desired_capabilities: desired capabilities as passed to the Appium webdriver
        :param implicit_wait: how long Appium will look for an element (you can look for an element before it is
                              rendered). Default 1 second
        :param appium_server: url of the appium server
        """
        # prepare options
        self.options = _get_android_default_options()
        self.options.udid = udid
        self.app_package = app_package
        if desired_capabilities:
            self.options.load_capabilities(desired_capabilities)

        logger.info("Connecting to Appium driver...")
        self.driver = _get_appium_driver(appium_server, udid, self.options)

        # the implicit wait time is how long appium looks for an element (you can try to find an element before it is rendered)
        self.implicit_wait = implicit_wait
        self.driver.implicitly_wait(implicit_wait)
        self.udid = self.driver.capabilities.get("udid")
        self.adb = AdbDevice(self.udid)

        # screen recorder
        self._screen_recorder = None

        logger.info(f"Activating package {self.app_package}...")
        self.activate_app()
        logger.info(f"App package {self.app_package} activated!")

    def activate_app(self):
        self.driver.activate_app(self.app_package)

    def terminate_app(self):
        self.driver.terminate_app(self.app_package)

    def restart_app(self):
        self.terminate_app()
        self.activate_app()

    def app_open(self) -> bool:
        return str(self.driver.current_package) == self.app_package

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._screen_recorder is not None:
            self.stop_recording_and_save_video()
        self.driver.__exit__(exc_type, exc_val, exc_tb)

    def back(self):
        self.driver.press_keycode(AndroidKey.BACK)

    def home(self):
        self.driver.press_keycode(AndroidKey.HOME)

    def open_notifications(self):
        self.driver.open_notifications()

    def scroll_to_find_element(self, resource_id: str = None, text_equals: str = None,
                               text_contains: str = None) -> WebElement:
        """
        This code will scroll in the current view until a certain element is found, and then return that element.
        The element can be searched for by resource-id and/or the text in that element.
        When defining the text of the element either an exact match or a textContains match can be used. These two can
        of course not be used at the same time.
        The method will keep scrolling until the element is found. If you look for something that doesn't exist you
        will have a bad time.
        The first matching element is returned when found.
        :param resource_id: the resource id of the element to look for.
        :param text_equals: the exact text of the element to look for. Cannot be used in combination with text_contains.
        :param text_contains: part of the text of the element to look for. Cannot be used in combination with
        text_equals.
        :return: The element when found.
        """
        if text_equals is not None and text_contains is not None:
            raise ValueError('text_equals and text_contains can not be used at the same time')
        if resource_id is None and text_contains is None and text_equals is None:
            raise ValueError('resource_id, text_equals and text_contains cannot all be None')

        resource_id_part = '' if resource_id is None else f'.resourceIdMatches("{resource_id}")'
        text_part = '' if text_equals is None else f'.text("{text_equals}")'
        text_contains_part = '' if text_contains is None else f'.textContains("{text_contains}")'

        java_code = f'new UiScrollable(new UiSelector().scrollable(true).instance(0)).scrollIntoView(new UiSelector(){resource_id_part}{text_part}{text_contains_part}.instance(0))'
        return self.driver.find_element(by=AppiumBy.ANDROID_UIAUTOMATOR, value=java_code)

    def swipe_to_find_element(self, xpath: str, max_swipes: int = 10) -> WebElement:
        for attempt in range(max_swipes):
            if self.is_present(xpath):
                return self.driver.find_element(AppiumBy.XPATH, xpath)
            else:
                # Element not found, perform swipe down
                print(f"Attempt {attempt + 1}: Element not found, swiping down")
                # Get the window size to determine the swipe dimensions
                window_size = self.driver.get_window_size()
                start_x = window_size['width'] / 2
                start_y = window_size['height'] * 0.8
                end_y = window_size['height'] * 0.2

                # Perform the swipe down
                self.driver.swipe(start_x, start_y, start_x, end_y, 500)

                # Wait a bit before the next attempt
                time.sleep(0.5)

        # If the loop completes, the element was not found
        raise NoSuchElementException(msg=f'After {max_swipes} swipes, cannot find element with xpath {xpath}')

    def is_present(self, xpath: str, implicit_wait: int = 0) -> bool:
        self.driver.implicitly_wait(implicit_wait)
        found = self.driver.find_elements(by=AppiumBy.XPATH, value=xpath)
        self.driver.implicitly_wait(self.implicit_wait)
        return len(found) > 0

    def start_recording(self, output_directory: str):
        if self._screen_recorder is None:
            self._screen_recorder_output_directory = output_directory
            self._screen_recorder = AdbScreenRecorder(self.adb)
            self._screen_recorder.start_recording()

    def stop_recording_and_save_video(self) -> [str]:
        if self._screen_recorder is None:
            return None
        video_files = self._screen_recorder.stop_recording(self._screen_recorder_output_directory)
        self._screen_recorder.__exit__(None, None, None)
        self._screen_recorder = None
        return video_files

    def new_screenshot_name(self):
        now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        device_name = self.options.device_name
        return Path(CACHE_FOLDER) / f'{now}-{device_name}-{uuid4()}.png'

    def find_text_ocr(self, text_to_find: str) -> [ocr.RecognizedText]:
        path = self.new_screenshot_name()
        screenshot_taken = False
        try:
            screenshot_taken = self.driver.get_screenshot_as_file(path)
            if not screenshot_taken:
                raise Exception(f'Screenshot could not be stored to {path}')
            found_text = ocr.find_text(str(path), text_to_find)
            return found_text
        finally:
            if screenshot_taken:
                os.remove(path)

    def click_text_ocr(self, text_to_click: str, click_first_when_multiple: bool = False):
        found_text = self.find_text_ocr(text_to_click)
        if len(found_text) == 0:
            msg = f'Could not find text {text_to_click} on screen so could not click it'
            log_error_and_raise_exception(logger, msg)
        if len(found_text) > 1:
            msg = f'Found multiple occurrences of text {text_to_click} on screen so could not determine what to click'
            if not click_first_when_multiple:
                log_error_and_raise_exception(logger, msg)
            else:
                logger.warning(f'Found multiple occurrences of text {text_to_click} on screen, clicking first one')
        x = found_text[0].bounding_box.middle[0]
        y = found_text[0].bounding_box.middle[1]
        self.driver.execute_script('mobile: clickGesture', {'x': x, 'y': y})

    def set_idle_timeout(self, timeout: int):
        # https://github.com/appium/appium-uiautomator2-driver#poor-elements-interaction-performance
        # https://github.com/appium/appium-uiautomator2-driver#settings-api
        settings = self.driver.get_settings()
        settings.update({"waitForIdleTimeout": timeout})
        self.driver.update_settings(settings)



from time import sleep
from typing import Dict, Optional

from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android.appium_actions import supported_version, AndroidAppiumActions

TELEGRAM_PACKAGE = 'org.telegram.messenger'
TELEGRAM_WEB_PACKAGE = 'org.telegram.messenger.web'


@supported_version("11.8.3")
class TelegramActions(AndroidAppiumActions):

    def __init__(self,
                 device_udid,
                 telegram_web_version: bool = False,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait=1,
                 appium_server='http://localhost:4723'):
        """
        Class with an API for Telegram Android using Appium. Can be used with an emulator or real device attached to the
        computer.
        This class can be used for both the Play Store version and the version found at telegram.org. When using the
        latter however, you need to use the `telegram_web_version` parameter, and set it to True.
        """
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      TELEGRAM_WEB_PACKAGE if telegram_web_version else TELEGRAM_PACKAGE,
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=implicit_wait,
                                      appium_server=appium_server)

    def _currently_at_homescreen(self, **kwargs) -> bool:
        return self.is_present('//android.widget.FrameLayout[@content-desc="New Message"]', **kwargs)

    def _currently_in_conversation(self, **kwargs) -> bool:
        return self.is_present('//android.widget.ImageView[@content-desc="Emoji, stickers, and GIFs"]', **kwargs)

    def _currently_in_camera(self, **kwargs) -> bool:
        return self.is_present('//android.widget.Button[lower-case(@content-desc)="shutter"]', **kwargs)

    def _currently_in_call(self, **kwargs) -> bool:
        return self.is_present('//android.widget.Button[lower-case(@text)="end call"]', **kwargs)

    def _currently_in_active_call(self, **kwargs) -> bool:
        """
        Returns true if currently in a call that is in actual progress (meaning: the other party answered the call).
        """
        return self.is_present('//android.widget.LinearLayout[@content-desc="Encryption key of this call"]', **kwargs)

    def _load_conversation_titles(self):
        while True:
            xpath = "//androidx.recyclerview.widget.RecyclerView/android.view.ViewGroup[not(@content-desc)]"
            convos = self.driver.find_elements(by=AppiumBy.XPATH, value=xpath)
            if len(convos) == 0:
                break
            print('some conversations still do not have a title loaded, clicking...')
            convos[0].click()
            self.driver.back()
            if not self._currently_at_homescreen():
                self.driver.back()
        print('all conversations loaded!')

    def return_to_homescreen(self, attempts: int = 10, try_restart: bool = True):
        """
        Returns to the start screen of Telegram
        :param attempts: Number of attempts to return to home screen. Avoids an infinite loop when a popup occurs.
        """
        attempt = 0
        while not self._currently_at_homescreen() and attempt < attempts:
            if self.driver.current_package != self.app_package:
                self.activate_app()
            self.driver.back()
            attempt += 1
        if not self._currently_at_homescreen():
            if try_restart:
                self.restart_app()
                self.return_to_homescreen(attempts, False)
            else:
                raise Exception('Tried to return to homescreen but ran out of attempts...')
        self._load_conversation_titles()
        sleep(1)

    def start_new_chat(self, chat: str):
        self.return_to_homescreen()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.FrameLayout[@content-desc="New Message"]').click()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.ImageButton[@content-desc="Search"]').click()
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.EditText[@text="Search"]').send_keys(chat)
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value=f'//android.view.ViewGroup[starts-with(lower-case(@text), "{chat.lower()}")]').click()

    def select_chat(self, chat: str):
        """
        Opens a given conversation based on the (partial) name of a chat.
        For groups or channels, it is advised to use :meth:`TelegramActions.select_group` or
        :meth:`TelegramActions.select_channel`, as the matching is more explicit
        :param chat: (part of) the conversation name to open
        """
        self.return_to_homescreen()
        xpath = f'//android.view.ViewGroup[starts-with(lower-case(@content-desc), "{chat.lower()}")]'
        self.driver.find_element(by=AppiumBy.XPATH, value=xpath).click()
        if not self._currently_in_conversation(implicit_wait=1):
            raise RuntimeError("Conversation was not opened after clicking the conversation")

    def select_group(self, group_name: str):
        """
        Opens a given conversation based on the exact name of a group.
        :param group_name: Exact group name to open
        """
        self.select_chat(f'Group. {group_name}')

    def select_channel(self, channel_name: str):
        """
        Opens a given conversation based on the exact name of a channel.
        :param channel_name: Exact channel name to open
        """
        self.select_chat(f'Channel. {channel_name}')

    def send_message(self, message: str, chat: str = None):
        """
        Send a message in the current or given chat
        :param message: The text message to send
        :param chat: Optional: The chat conversation in which to send this message, if not currently in the desired chat
        """
        self._if_chat_go_to_chat(chat)
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.EditText[@text="Message"]').send_keys(
            message)

        # The actual send button is not in the same place as the element. The button is at about 75% of the box.
        # We adjust the location of the click from the middle to the right of the box.
        location = self._find_button_location(0.75, 0.5, '//android.view.View[@content-desc="Send"]')
        self.driver.tap([(location)])

    def reply_to_message(self, message_to_reply_to: str, reply: str, chat: str = None):
        """
        Send a text message as a reply to a previous message.
        This method will scroll up until the given message text is found, select that message, and reply to it
        :param message_to_reply_to: the message to reply to
        :param reply: The text message that will be sent as a reply
        :param chat: Optional: The chat conversation in which to send this message, if not currently in the desired chat
        """
        try:
            self._tap_message(message_to_reply_to, chat)
        except:
            raise Exception(f'Failed to scroll to message "{message_to_reply_to}", does it exist?')
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.TextView[@text="Reply"]').click()
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.EditText').send_keys(reply)
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.view.View[@content-desc="Send"]').click()

    def emoji_reply_to_message(self, message_to_respond_to: str, emoji_to_respond_with: str = '👍', chat: str = None):
        """
        Send emoji response to a previous message.
        This method will scroll up until the given message text is found, select that message, and reply to it
        :param message_to_reply_to: the message to reply to
        :param emoji_to_respond_with: the emoji with which to respond. Optional, by default '👍'
        :param chat: Optional: The chat conversation in which to send this message, if not currently in the desired chat
        """
        self._tap_message(message_to_respond_to, chat)
        if self.is_present(f'//android.widget.FrameLayout[@text="{emoji_to_respond_with}"]', implicit_wait=2):
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value=f'//android.widget.FrameLayout[@text="{emoji_to_respond_with}"]').click()
        else:
            print(f'ERROR: emoji {emoji_to_respond_with} does not exist on screen!')
            self.driver.back()

    def _tap_message(self, message_to_tap, chat):
        self._if_chat_go_to_chat(chat)
        message = self.scroll_to_find_element(text_contains=message_to_tap)
        # we want to click the middle right part of the message to open the menu
        x = message.location['x'] + message.size['width'] - 50
        y = message.location['y'] + message.size['height'] / 2
        print(f'Tapping screen at ({x},{y})')
        self.driver.tap([(x, y)])

    def take_and_send_picture(self,
                              chat: str = None,
                              caption: str = None,
                              wait_time: int = 1,
                              front_camera: bool = False):
        """
        Opens the embedded camera in the Telegram app, takes a picture, and sends it.
        :param chat: Optional: the conversation in which to send the picture. If not used, we assume a conversation is opened
        :param caption: Optional: a caption to include with the picture
        :param wait_time: Optional: time to wait (in seconds) after opening the camera before taking a picture. Default 1s
        :param front_camera: Optional: whether to use the front camera. Default False
        """
        self._if_chat_go_to_chat(chat)
        # click the attachment icon
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.ImageView[lower-case(@content-desc)="attach media"]').click()

        camera_preview_xpath = '//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout[2]'
        shutter_xpath = '//android.widget.Button[lower-case(@content-desc)="shutter"]'
        for i in range(10):
            if self.is_present(camera_preview_xpath):
                self.driver.find_element(by=AppiumBy.XPATH, value=camera_preview_xpath).click()
                if self.is_present(shutter_xpath):
                    break
            sleep(1)
        if not self.is_present(shutter_xpath):
            raise Exception("Clicking the camera preview to take a picture failed.")

        # switch camera if need be
        if front_camera:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.ImageView[lower-case(@content-desc)="switch camera"]').click()
        sleep(wait_time)
        # take picture
        self.driver.find_element(by=AppiumBy.XPATH, value=shutter_xpath).click()
        # add caption
        if caption is not None:
            text_field = '//android.widget.EditText'
            for i in range(10):
                if self.is_present(text_field):
                    break
                sleep(1)
            self.driver.find_element(by=AppiumBy.XPATH, value=text_field).click()
            self.driver.find_element(by=AppiumBy.XPATH, value=text_field).send_keys(caption)

            if self.driver.is_keyboard_shown():
                self.driver.find_element(by=AppiumBy.XPATH,
                                         value='//android.widget.EditText/../../android.widget.ImageView').click()
        # press send
        # The actual send button is not in the same place as the element. The button is at about 75% of the box.
        # We adjust the location of the click from the middle to the right bottom corner of the box.
        location = self._find_button_location(0.75, 0.75, '//*[lower-case(@content-desc)="send"]')
        self.driver.tap([(location)])

        sleep(0.3)  # the animation after sending a picture might throw off the script

    def start_call(self, chat: str = None, video: bool = False) -> bool:
        """
        Makes a call and ends the call after a given number of seconds.
        The call is made to either
            * the current conversation we're in
            * a given contact name
        :param chat: Optional: name of the conversation to start a call in
        :param video: False (default) for voice call, True for video call.
        :return: True if call was successfully started
        """
        self._if_chat_go_to_chat(chat)
        if not video:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.ImageButton[lower-case(@content-desc)="call"]').click()
        else:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.ImageButton[lower-case(@content-desc)="more options"]').click()
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.TextView[lower-case(@text)="video call"]').click()
        # wait a short while (max 2s) for the call to have started
        return self._currently_in_call(implicit_wait=2)

    def get_call_status(self) -> Optional[str]:
        """
        Returns a string describing the status of the current call, which is a text visible on the call screen.
        If the call is in progress, no text indicates this, but this method will return "In progress".
        If not currently in a call, this method returns None.
        Known statuses include 'Requesting', 'Waiting' and 'Ringing'.
        """
        if not self._currently_in_call():
            return None
        if self._currently_in_active_call():
            return "In progress"
        status_element = '//android.widget.LinearLayout[ends-with(@text, "Telegram Call")]/android.widget.FrameLayout/android.widget.TextView'
        status_element = self.driver.find_element(by=AppiumBy.XPATH, value=status_element)
        return status_element.get_attribute("text")

    def end_call(self):
        """
        Ends the current call. Assumes the call screen is open.
        """
        if not self._currently_in_call():
            raise Exception('Cannot end call while not in call screen.')
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.Button[lower-case(@text)="end call"]').click()

    def answer_call(self):
        """
        Answer when receiving a call via Telegram.
        """
        self.driver.open_notifications()
        if not self.is_present('//android.widget.Button[lower-case(@content-desc)="answer"]', implicit_wait=3):
            self.driver.back()
            raise Exception('Tried to answer call, but couldn\'t find notification to answer any calls')
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.Button[lower-case(@content-desc)="answer"]').click()
        open_call_button = '//android.widget.TextView[@resource-id="android:id/title" and @text="Ongoing Telegram call"]'
        if self.is_present(open_call_button):
            self.driver.find_element(by=AppiumBy.XPATH, value=open_call_button).click()

    def decline_call(self):
        """
        Declines an incoming Telegram call.
        """
        self.driver.open_notifications()
        if not self.is_present('//android.widget.Button[lower-case(@content-desc)="decline"]', implicit_wait=3):
            self.driver.back()
            raise Exception('Tried to answer call, but couldn\'t find notification to answer any calls')
        self.driver.find_element(by=AppiumBy.XPATH,
                                 value='//android.widget.Button[lower-case(@content-desc)="decline"]').click()
        self.driver.back()

    def toggle_video_in_call(self):
        """
        Toggles video in an ongoing call.
        """
        if not self._currently_in_call():
            raise Exception('Expected to be in a call, but could not detect call screen!')
        if self.is_present('//android.widget.TextView[lower-case(@text)="start video"]'):
            # turn on video:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.FrameLayout[@content-desc="Start Video"]').click()
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.TextView[@content-desc="Share Video"]').click()
        else:
            self.driver.find_element(by=AppiumBy.XPATH,
                                     value='//android.widget.FrameLayout[@content-desc="Stop Video"]').click()

    def flip_video_in_call(self):
        """
        Switches between the front camera and rear camera while in a video call.
        """
        if not self._currently_in_call():
            raise Exception('Expected to be in a call, but could not detect call screen!')
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.FrameLayout[@content-desc="Flip"]').click()

    def _if_chat_go_to_chat(self, chat: str):
        if chat is not None:
            self.select_chat(chat)
            sleep(1)
        if not self._currently_in_conversation():
            raise Exception('Expected to be in conversation screen now, but screen contents are unknown')

    def _find_button_location(self, width_ratio: float, height_ratio: float, xpath: str):
        send_button = self.driver.find_element(by=AppiumBy.XPATH,
                                               value=xpath)
        top_left = send_button.location['x'], send_button.location['y']
        size = send_button.size['height'], send_button.size['width']
        location = int(top_left[0] + width_ratio * size[1]), int(top_left[1] + height_ratio * size[0])
        return location
